import numpy as np
import pandas as pd
import os

# due to a known issue with memory leak on Windows with MKL, set the following:
# os.environ["OMP_NUM_THREADS"] = "1"
# os.environ["MKL_NUM_THREADS"] = "1"
import warnings
from sklearn.metrics.pairwise import cosine_distances
from sklearn.manifold import MDS  # , TSNE
from sklearn.decomposition import PCA
from sklearn.cluster import KMeans
from sklearn.neighbors import KDTree
from sklearn.utils import Bunch
import sksurv

# from .utilities import  frmt_pretty_print
from .tree_representation_utils import tree_splits_to_vector  # tree_vector
from .tree_representation_utils import rule_splits_to_vector, add_emergency_noise
from .utilities import predict_helper, safe_element_to_scalar


class TreeExtraction:  # is it convenient if it inherits?

    RAND_SEED = 0
    BINARY_KEYS = ["bin", "binary"]
    SURVIVAL_KEYS = ["surv", "survival"]
    REGRESSION_KEYS = ["regress", "regression", "regr"]
    MTC_KEYS = ["multi-l", "multi-label", "mtc", "multi"]
    MTR_KEYS = ["multi-t", "multi-target", "mtr"]
    MSA_KEYS = ["multi-sa", "multi-variate-sa", "mvsa"]

    def __init__(
        self,
        proj_method,
        dissim_method,
        feature_represent,
        n_trees,
        n_dims,
        n_clusters,
        pre_select_loss,
        fidelity_measure,
        clf,
        oracle_sample,
        set_up,
        sample,
        verbose,
        output_explain=False,
    ):
        # consider inheriting from LocalMethod, smth like:
        # LocalMethod.__init__(self) (??)
        self.proj_method = proj_method
        self.dissim_method = dissim_method
        self.feature_represent = feature_represent
        self.n_trees = n_trees  # repetitive... correct?
        self.n_dims = n_dims
        self.n_clusters = n_clusters
        self.pre_select_loss = pre_select_loss  # drop?
        self.fidelity_measure = fidelity_measure
        self.clf = clf
        self.oracle_sample = oracle_sample
        self.set_up = set_up
        self.sample = sample  #  X[idx:idx+1]
        self.verbose = verbose  # to improve (set levels for example)
        self.final_trees_idx = None  # non extracted yet
        self.cluster_sizes = None  # non extracted yet
        self.output_explain = output_explain

    # adaptation of n_trees to case where the proportion is given instead, is
    # is handled in the .fit of LocalMethod

    def get_params(self, deep=True):  # set deep=True if it does not work
        return {
            "dissim_method": self.dissim_method,
            "proj_method": self.proj_method,
            "n_trees": self.n_trees,
            "n_dims": self.n_dims,
            "n_clusters": self.n_clusters,
        }

    def set_params(self, **params):
        for key, value in params.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                raise ValueError(f"Invalid parameter: {key}")
        return self

    def main_fit(self):
        """
        This is the main function, does the following:
            raw_tree data:
            - compute tree loss, (and sort for pre-filtering)
            - computes vectos_to_matrix,
            - performs dim_reduction step
            extract_final_trees
            - clusters the output tree-representations and select centers ( with weights)

        the .main_fit summarises the whole method's procedure,
        the steps are:
            - tree_matrix is build, and tree "local loss"(1-faithfulness)
            is calculated
            - pre_selection_trees phase selects top \\lambda trees ( tunable)
            - dim_reduction performs MDS to reduce dimensionality (tunable)
            - cluster_and-extract performs K-means and selects trees closest
            to each cluster center, also outputs cluster sizes
            - local predict performs the weighted avergae of the extracted
            trees, based on the inputted cluster sizes
            - finally the clustering performance is evaluated with the .score
            method. Needed for tuning and for reporting.

        HL bunch properties:
            Bunch(proj_data=proj_trees, ---> the tree representation, projected
                  matrix=tree_matrix,   ---> the original tree representation (kept for debug?)
                  index=selected_trees_idx,   ---> indeces of the orginal tree learners
                                            (needed as the order is lost ater pre-selection step)
                  rf_pred=rf_pred,      ---> RF original prediction
                  loss=selected_losses)       ---> distance ( loss) to full RF prediction
        """
        kmeans, selected_bunch = self.preselect_represent_cluster_trees()

        self.final_trees_idx, self.cluster_sizes = self.extract_final_trees(
            selected_bunch.proj_data, selected_bunch.index, kmeans
        )
        return self

    def preselect_represent_cluster_trees(self):
        """
        transform trees to vectors and store everything in matrix,
        the output is matrix of the form (n_selected_trees, n_features),
        indeces of the original trees are also stored through the DataFrame
        """
        ### pre-selection step:
        tree_local_losses, rf_pred = self.calcul_tree_proximity_loss(
            self.sample
        )  # w.r.t. to sample we want to explain

        selected_trees_idx = np.argsort(tree_local_losses)[: self.n_trees]  # sort trees losses
        selected_preds = [
            predict_helper(self.clf[i], self.sample.values) for i in selected_trees_idx
        ]

        selected_losses = np.array([tree_local_losses[i] for i in selected_trees_idx])

        dict_trees = {}  # empty dictionary to fill: {index : tree-vector}

        if self.dissim_method == "trees":
            for idx in selected_trees_idx:
                dict_trees.update(
                    {idx: tree_splits_to_vector(self.clf, idx, self.feature_represent)}
                )
            tree_matrix = pd.DataFrame.from_dict(dict_trees, "index")

        elif self.dissim_method == "rules":
            for idx in selected_trees_idx:
                dict_trees.update(
                    {idx: rule_splits_to_vector(self.clf, idx, self.feature_represent, self.sample)}
                )
            tree_matrix = pd.DataFrame.from_dict(dict_trees, "index")

        # for PCA, the tree_representation is enough: tree_vector / rule_vector,
        # therefore, no pairwise distance is needed

        if self.proj_method == "PCA":
            pass  # no pairwise distance calculation

        # for MDS, a transformation to distance square matrix is needed first
        elif self.proj_method == "MDS":

            tree_matrix = self.transform_to_symm_matrix(tree_matrix)
            # in case distances all collapsed to 0
            if np.abs(tree_matrix).max() < 1e-6:
                tree_matrix = add_emergency_noise(tree_matrix)

        else:
            raise KeyError(f"projection method '{self.proj_method}' not recognised.")

        # by now, the tree_matrix has been computed and is in the appropriate format (symmetric)

        # dim_reduction inherits self.n_dims and self.proj_method
        # dim_reduction calls either PCA or MDS, according to proj-method variable
        proj_trees = self.dim_reduction(
            tree_matrix
        )  # from pandas to numpy. Indexes stored in Bunch

        # clustering step (indices retained in Bunch object)
        selected_bunch = Bunch(
            proj_data=proj_trees,
            matrix=tree_matrix,
            index=selected_trees_idx,
            loss=selected_losses,
            pred=selected_preds,
            rf_pred=rf_pred,
            set_up=self.set_up,
        )
        # numpy 2d-array (lambda,2), (lambda, lambda), orig. indices, losses
        if self.n_clusters > 1:  # K-Means is performed on the projected trees
            with (
                warnings.catch_warnings()
            ):  # ignore known issue on scikit-learn (as of scikit-learn 1.2.2)
                warnings.filterwarnings("ignore", message=".*memory leak on Windows with MKL*")
                # kmeans = None # initialise with dummy line (why? is the filterswarning the problem?)
                kmeans = KMeans(
                    n_clusters=self.n_clusters,
                    init="k-means++",
                    n_init=5,
                    max_iter=1000,
                    random_state=TreeExtraction.RAND_SEED,
                ).fit(proj_trees)

        else:  # avoid running K-Means (raises warning)
            # output info in the same format so that the cases are indistinguishable
            one_label = np.zeros(proj_trees.shape[0], dtype=np.int32)
            one_center = np.mean(proj_trees, axis=0)
            # mimic a kmeans instance
            kmeans = Bunch(
                labels_=one_label,  # kmeans.lables_ is a 1D array of 1s only (single class)
                cluster_centers_=np.array([one_center]),  # kmeans.centers_ is a 2D array
            )

        return kmeans, selected_bunch

    def calcul_tree_proximity_loss(self, sample):

        # get number of output labels
        if self.set_up not in self.SURVIVAL_KEYS:
            n_real_outputs = self.clf.n_outputs_
        else:
            n_real_outputs = 1
        if self.set_up in self.MSA_KEYS:
            warnings.warn("If newly implemented, check the self.clf.n_outputs_ part of this code.")

        # output vector consistent with utilities.predict_helper function output
        if n_real_outputs > 1:  # multi-output set-up
            tree_preds = np.zeros([self.clf.n_estimators, n_real_outputs])
        else:  # single output set-up (binary, survival, regression)
            tree_preds = np.zeros(self.clf.n_estimators)

        my_pre_select_loss = np.zeros(self.clf.n_estimators)

        # GOAL HERE is to store tree predictions in a consistent format.
        # Later formatting is performed with other functions

        rf_pred = predict_helper(self.clf, sample)
        for k in range(self.clf.n_estimators):
            tree_preds[k] = safe_element_to_scalar(predict_helper(self.clf[k], sample.values))

        # tree loss (euclidean norm of the predition vector vs real target(s))
        if self.pre_select_loss in ["fidel-L2", "L2"]:
            for k in range(self.clf.n_estimators):
                my_pre_select_loss[k] = np.linalg.norm(rf_pred - tree_preds[k])
        elif self.pre_select_loss in ["fidel-cosine", "cosine", "cos"]:
            my_pre_select_loss[k] = cosine_distances(rf_pred, tree_preds[k])
        else:
            raise KeyError("Pre-selection criterion '{}' not recognized")

        return my_pre_select_loss, rf_pred

    def tree_prediction(self, clf_tree):  # (self, selected_indeces, sample):

        # same as before: clf_tree does not inherit feature_names_in_ correctly

        if hasattr(clf_tree, "predict_proba"):
            if clf_tree.n_outputs_ > 1:  # multi-label-classification case
                single_tree_pred = np.array(clf_tree.predict_proba(self.sample.values))[:, :, 1]
            else:  # binary case
                single_tree_pred = clf_tree.predict_proba(self.sample.values)[:, 1][0]
        elif hasattr(clf_tree, "predict"):  # regression, survival, multi-target regression
            if (
                isinstance(clf_tree, sksurv.tree.SurvivalTree) or clf_tree.n_outputs_ == 1
            ):  # survival OR regression
                single_tree_pred = float(clf_tree.predict(self.sample.values))

            else:  # multi-target regression
                # nothing is accurate here, just random
                single_tree_pred = clf_tree.predict(self.sample.values)

        else:
            raise ValueError("No case was found for clf_tree, check your input")

        return single_tree_pred

    def jaccard_pair_distance(self, tree_1, tree_2):
        return np.sum(np.minimum(tree_1, tree_2)) / np.sum(np.maximum(tree_1, tree_2))

    def transform_to_symm_matrix(self, matrix_df_input):
        # assumes the input is a DataFrame (the DataFrame is create just before this
        # function is called, so I wouldn't even bother checking for type
        matrix_input = matrix_df_input.values  # to np.ndarray

        # assumes each ROW corresponds to a representation of a tree
        size = matrix_input.shape[0]
        A = np.zeros([size, size])  # this part can be optmisied (and double checked)

        for i in range(size):
            for j in range(i + 1):
                A[i, j] = 1 - self.jaccard_pair_distance(
                    matrix_input[i], matrix_input[j]
                )  # 1-Jacc for dissimilarity
                A[j, i] = A[i, j]
        return A  # no need to store DataFrame, the original indeces
        # are stored in some selected_indeces and in the soon to appear selected_bunch object
        # return pd.DataFrame(A, columns=matrix_df_input.columns,
        #                    index=matrix_df_input.columns)

    def pre_selection_trees(self, tree_matrix):
        tree_prox_loss = tree_matrix.loss_loss
        selected_trees_idx = np.argsort(tree_prox_loss)[: self.n_trees]  # first sort all trees

        selected_tree_losses = np.sort(tree_prox_loss)[: self.n_trees]  # get LOWEST loss
        selected_trees = tree_matrix.matrix.reindex(index=selected_trees_idx).reindex(
            columns=selected_trees_idx
        )
        selected_bunch = Bunch(trees=selected_trees, loss=selected_tree_losses)
        return selected_bunch

    def dim_reduction(self, selected_trees):

        # setting up "true" number of output dimensions, furtrmore
        # prevent errors from raising if n_dims > n_samples
        if self.n_dims is None:  # useful for MDS only (PCA can skip this part)
            true_dims = min(selected_trees.shape[0], selected_trees.shape[1])  # keep all dimensions
        elif self.n_dims >= 1:  # should not be needed, but sometimes it is...
            true_dims = min(self.n_dims, selected_trees.shape[0], selected_trees.shape[1])
        else:
            true_dims = (
                self.n_dims
            )  # follows sklearn's implementation (amount of explained variance)

        # apply the dim reduciton method from the candidate ETrees object
        if self.proj_method == "PCA" and self.n_dims is not None:  # if None, skip completely
            m = PCA(n_components=true_dims, random_state=TreeExtraction.RAND_SEED)
            proj_trees = m.fit_transform(selected_trees)

        elif self.proj_method == "MDS":
            m = MDS(
                n_components=true_dims,
                metric=True,
                max_iter=1000,  # metric=True, sure?
                random_state=TreeExtraction.RAND_SEED,
                dissimilarity="precomputed",
            )
            proj_trees = m.fit_transform(selected_trees)

        else:  # method == PCA and n_dims == None: do nothing :-P
            proj_trees = selected_trees

        return proj_trees

    def extract_final_trees(self, proj_trees_data, origin_indeces, kmeans):

        cluster_sizes = np.bincount(kmeans.labels_)
        search_tree = KDTree(proj_trees_data)
        final_trees = search_tree.query(
            kmeans.cluster_centers_, k=1, return_distance=False
        ).flatten()
        # flattening is necessay, otherwise 2D array is outputted
        # search_tree.query returns: (dist, indeces loc), we are interested in indeces only
        # however, pandas indeces are now lost, we need to retrieve them:
        # self.final_trees_idx= proj_trees.index[final_trees].values
        self.final_trees_idx = [origin_indeces[k] for k in final_trees]
        # TreeExtraction.plot_my_clusters(self, proj_trees, kmeans)
        return self.final_trees_idx, cluster_sizes  # it's an array, right?

    def local_prediction(self):  # self clf, X, sample,

        assert np.shape(self.cluster_sizes)[0] == np.shape(self.final_trees_idx)[0]

        # get number of output labels
        n_real_outputs = self.clf.n_outputs_

        # TODO include case for multi-output SA when available
        if self.set_up in self.SURVIVAL_KEYS:
            n_real_outputs = 1
            # self.clf.n_outputs_  on RSF gives shape: (unique_times_, ) which is miselading

        # output vector consistent with utilities.predict_helper function output
        if n_real_outputs > 1:  # multi-output set-up
            btrex_pred = np.zeros([1, n_real_outputs])
        else:  # single output set-up (binary, survival, regression)
            btrex_pred = np.zeros(1)

        for t, cluster_size in zip(self.final_trees_idx, self.cluster_sizes):
            cluster_weight = cluster_size / np.sum(self.cluster_sizes)
            # assert isinstance(predict_helper(self.clf[t], self.sample.values), (float, int))
            btrex_pred += predict_helper(self.clf[t], self.sample.values) * cluster_weight

        return btrex_pred

    ### HERE ADD ORACLE PREDICTION IN THE FUTURE (decouple from RF)
    def oracle_prediction(self):  # add, inherit prediction method

        return predict_helper(self.clf, self.sample)

    def score(self, fidelity_measure, oracle_sample):
        y_local = self.local_prediction()
        y_pred = (
            self.oracle_prediction() if oracle_sample is None else oracle_sample
        )  # not always same format as y_local

        # y_local and y_pred might not have the exact same format,
        # but the scoring measures still work (as for now)

        if fidelity_measure in ["fidel-L2", "L2"]:
            return 1 - np.linalg.norm(y_pred - y_local)  # 1 - L2 distance
        elif fidelity_measure in ["fidel-cosine", "cosine", "cos"]:
            # it outputs a matrix of distances, we are interested in a 1x1 comparison
            return 1 - float(cosine_distances(y_pred.reshape(1, -1), y_local.reshape(1, -1)))
        else:
            raise KeyError(f"Input {self.fidelity_measure} is not recognised")

    #### this code is of poor quality, seldom used ####
    # TODO clean up from this point onwards
    # def print_rules(self, save_name_folder, i, y_true=np.nan):

    #     # the name of the folder will be given as an input to the class method
    #     print("Calling print_rules now")

    #     rf = self.clf
    #     sample = self.sample #pd Series
    #     if hasattr(rf, "feature_names_in_"):
    #         feature_names = rf.feature_names_in_
    #     else:
    #         feature_names = [f"X{i}" for i in range(rf.n_features_in_)]

    #     weights_dict = dict(zip(self.final_trees_idx,
    #                             self.cluster_sizes/np.sum(self.cluster_sizes)))

    #     sample_array = self.sample.to_numpy().reshape(1,-1).astype("float32")

    #     sample_string = "Sample" + str(i) + "_input.txt"
    #     full_save_sample = os.path.join(save_name_folder, "path-print", sample_string)

    #     dir_save_tree = os.path.join(save_name_folder, "path-print")
    #     dir_plot_tree = os.path.join(save_name_folder, "plot-save")

    #     directories = [dir_save_tree, dir_plot_tree]
    #     for new_dir in directories:
    #         if not os.path.exists(new_dir): #if the data folder does not exists (e.g. new dataset being used)
    #             os.makedirs(new_dir) # then create the new folder

    #     #plotting covariate values of the sample to be explained
    #     with open(full_save_sample, 'w+') as f:
    #         f.write("###### SAMPLE to explain ######\n")
    #         for feat in feature_names:
    #             sample_value = sample[feat].values[0]
    #             f.write("{:13}: {:7} \n".format(feat, str(sample_value)))

    #     for tree_idx in self.final_trees_idx:

    #         tree_path_string = "Sample" + str(i) + "_Tree" + str(tree_idx) + "_Tree.txt"
    #         final_leaf_string = "Sample" + str(i) + "_Tree" + str(tree_idx) + "_Leaf.txt"
    #         final_plot_string = "Sample" + str(i) + "_Tree" + str(tree_idx) + "_Plot.png"

    #         full_save_tree = os.path.join(dir_save_tree, tree_path_string)
    #         full_save_leaf = os.path.join(dir_save_tree, final_leaf_string)
    #         full_save_plot = os.path.join(dir_plot_tree, final_plot_string)

    #         DT_path = self.clf[tree_idx].tree_.decision_path(sample_array).toarray().flatten()

    #         #if isinstance(rf[tree_idx], sklearn.tree._classes.DecisionTreeClassifier):
    #         if self.set_up.lower() in self.BINARY_KEYS:
    #             leaf_print = float(rf[tree_idx].predict_proba(sample)[:,1])
    #             rf_pred = float(rf.predict_proba(sample)[:,1])
    #             rf_preds = [float(rf[i].predict_proba(sample)[:,1])
    #                         for i in range(rf.n_estimators)]
    #         elif self.set_up.lower() in self.MTC_KEYS:
    #             # not sure the formatting is OK (probably not)
    #             leaf_print = rf[tree_idx].predict_proba(sample)[:,1].ravel()
    #             rf_pred = rf.predict_proba(sample)[:,1].ravel()

    #         elif self.set_up.lower() in self.REGRESSION_KEYS:
    #         #elif isinstance(rf[tree_idx], sklearn.tree._classes.DecisionTreeRegressor):
    #             leaf_print = rf[tree_idx].predict(sample).ravel()
    #             rf_pred = float(rf.predict(sample)[:,1])
    #             rf_preds = [float(rf[i].predict(sample)[:,1])
    #                         for i in range(rf.n_estimators)]

    #         elif self.set_up.lower() in self.SURVIVAL_KEYS: # is survival case
    #             leaf_print = rf[tree_idx].predict(sample).ravel() # risk score
    #             leaf_surv_curve = rf[tree_idx].predict_survival_function([sample], return_array=False)

    #             rf_pred = rf.predict(sample).ravel()
    #             rf_preds = [rf[i].predict(sample)[:,1].ravel()
    #                         for i in range(rf.n_estimators)]

    #         else:
    #             raise KeyError("Set-up {} is not recognised".format(self.set_up))

    #         tree_structure = rf[tree_idx].tree_

    #         intervals = {feat : [-np.inf, np.inf] for feat in feature_names}

    #         # this is weird, why does sample change type all of the sudden?
    #         if isinstance(sample, pd.Series):
    #             sample = sample.to_numpy().reshape(-1,1).flatten() #from single column to single line
    #         elif isinstance(sample, np.ndarray):
    #             sample = sample.reshape(-1,1).flatten()

    #         if self.output_explain == False:
    #             print("output_explain is \'False\', no output files are generated")

    #         def current_pred(tree_struct, node_id, indent, set_up, opened_f):
    #             if set_up.lower() in ["bin", "binary"]:
    #                 pred_values = tree_struct.value[node_id][0,:]
    #                 avg_pred = pred_values[1]/np.sum(pred_values)
    #                 if node_id == 0:
    #                     opened_f.write("base pred:{} {:#4g}\n".format(indent, avg_pred))
    #                 else:
    #                     opened_f.write("curr pred:{} {:#4g}\n".format(indent, avg_pred))

    #             else:
    #                 raise Warning("Current node prediction is not yet implemented for this scenario")

    #         if self.output_explain == True:

    #             sample_np = sample.values.T.ravel()

    #             with open(full_save_tree, 'w+') as f:

    #                 # print decision path and store intervals that identify the final leaf

    #                 def recurse(node, depth, sample_np, intervals):

    #                     indent = "   " * depth
    #                     if tree_structure.feature[node] != -2: # is not a leaf #_tree.TREE_UNDEFINED:
    #                         name = feature_names[tree_structure.feature[node]]
    #                         threshold = tree_structure.threshold[node]

    #                         if DT_path[node] == 1 and sample_np[tree_structure.feature[node]] <= threshold:
    #                             intervals[name][1] = threshold # reduce feature upper bound
    #                             # avoid priting this repreatedly  during recursion
    #                             DT_path[node] = 0
    #                             if self.set_up.lower() in ["bin", "binary"]:
    #                                 current_pred(tree_structure, node, indent, self.set_up, f)
    #                             f.write("node.{:4}: {}if {:8} <= {:#4g}\n".format(node, indent, name, threshold))
    #                             # works only for binary case

    #                         recurse(tree_structure.children_left[node], depth + 1, sample_np, intervals)

    #                         if DT_path[node] == 1 and sample_np[tree_structure.feature[node]] > threshold:
    #                             intervals[name][0] = threshold # increase feature lower bound
    #                             DT_path[node] = 0
    #                             if self.set_up.lower() in ["bin", "binary"]:
    #                                 current_pred(tree_structure, node, indent, self.set_up, f)
    #                             f.write("node.{:4}: {}if {} > {:#4g}\n".format(node, indent, name, threshold))
    #                             #print("node.{:4}: {}if {:8} <= {:5}\n".format(node, indent, name, threshold))
    #                         recurse(tree_structure.children_right[node], depth + 1, sample_np, intervals)
    #                     else: #it is undefined, it is therefore a leaf (?)
    #                         if DT_path[node] == 1:
    #                             # no need to print twice with current_pred function
    #                             f.write("leaf.{:4}: {}return {:#4g}\n".format(node, indent, leaf_print))

    #                             #print("predicted:{}\n".format(leaf_print))
    #                             f.write("###################\n")
    #                             f.write("Tree weight:   {:.2f}\n".format(weights_dict[tree_idx]))
    #                             f.write("Tree predicts: {:.4f}\n".format(frmt_pretty_print(leaf_print)))
    #                             f.write("RF prediction: {:.4f}\n".format(frmt_pretty_print(rf_pred)))
    #                             f.write("True label: {}".format(y_true))

    #                 recurse(0, 1, sample_np, intervals) # problem arises here

    #                 if self.set_up.lower() in ["survival", "surv"]:
    #                     plt.figure()
    #                     #plt.title("Tree {} predicting sample {}".format()) not available at this level, it's in main!!
    #                     plt.plot(leaf_surv_curve[0].x, leaf_surv_curve[0].y)
    #                     plt.ylim(ymin=0, ymax=1)
    #                     plt.savefig(full_save_plot)
    #                     plt.legend()
    #                     plt.show()

    #                 if self.verbose >= 5.0:
    #                     pd.Series(rf_preds).plot.density(color='blue', label="RF predictions")
    #                     plt.title('RF prediction distribution')
    #                     plt.axvline(x=rf_pred, label="y pred.")
    #                     #if self.set_up.lower() in ["regress", "regr", "regression"]:
    #                     plt.axvline(x=y_true, label="y true", color="green")
    #                     plt.xlim(0,1)
    #                     plt.legend()
    #                     #plt.savefig(full_save_plot)
    #                     plt.show()

    #             with open(full_save_leaf, 'w') as f:
    #                 f.write("###### considered covariates ######\n")
    #                 for j, k in zip(feature_names, range(len(feature_names))):
    #                     if intervals[j][0] > -np.inf or intervals[j][1] < np.inf:
    #                         f.write("{:10}: {:#4g} \n".format(str(j), sample_np[k]))
    #                 f.write("#### True label: {} ####".format(y_true))
    #                 f.write("\n###### final intervals ########\n")

    #                 for item in intervals:
    #                     if intervals[item][0] > -np.inf or intervals[item][1] < np.inf:
    #                         f.write("{:#4g} < {:16} <= {:#4g} \n".format(intervals[item][0],
    #                                             str(item).center(16), intervals[item][1]))

    #                 with open(full_save_tree) as f: #printing tree-rule structure on console
    #                     print(f.read())

    #                 with open(full_save_leaf) as f:
    #                     print(f.read()) #printing (simplified) leaf structure on console

    # return tree_rules, leaf_struct



import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import gaussian_kde
from matplotlib.patches import FancyArrowPatch
from .utilities import frmt_pretty_print


def _input_validation(rules, preds, baselines, weights):
    """
    Validates the input parameters for consistency.

    Raises:
    - AssertionError: If the lengths of `rules`, `preds`, `baselines`, and `weights` are not equal.
    """
    assert (
        len(rules) == len(preds) == len(baselines) == len(weights)
    ), f"All input lists must have the same lengths, found {len(rules)}, {len(preds)}, {len(baselines)}, {len(weights)}."
    return None


def compute_max_visual_len(rules, max_rulelen=None):

    # Calculate the maximum rule length from the list of rules.
    max_rulelen_calculated = max(len(rule) for rule in rules)

    # Use the provided max_rulelen if it is not None and less than the calculated value.
    if max_rulelen is not None:
        max_rulelen = min(max_rulelen_calculated, max_rulelen)
    else:
        max_rulelen = max_rulelen_calculated

    return max_rulelen


def define_relative_position(n_cols):

    assert isinstance(n_cols, int), f"n_cols must be int, found {type(n_cols)} instead"

    if n_cols == 1:
        pos_list = ["center"]
    elif n_cols == 2:
        pos_list = ["left", "right"]
    elif n_cols > 2:
        pos_list = n_cols * ["center"]
        pos_list[0] = "left"
        pos_list[-1] = "right"
    else:
        raise ValueError(f"n_cols but be positive integer, got {n_cols} instead")
    return pos_list


def convert_to_data_coords(ax, point):
    """
    Converts a point from axes fraction coordinates to data coordinates.
    """
    x_range = ax.get_xlim()
    y_range = ax.get_ylim()

    x_data = x_range[0] + point[0] * (x_range[1] - x_range[0])
    y_data = y_range[0] + point[1] * (y_range[1] - y_range[0])

    return (x_data, y_data)


def plot_arrow(ax, pos, weight, pred_out, fontsize, tot_digits):

    # Define start position of the arrow based on predicted value (min-max transformed)
    x_min, x_max = ax.get_xlim()
    x_start = (pred_out - x_min) / (x_max - x_min)
    ax.axis("off")
    y_start = 0.98  # Starting height of the arrow

    if pos in ["c", "center"]:
        x_end, y_end = 0.5, 0.05

    elif pos in ["r", "right"]:
        x_end, y_end = -0.02, 0.05

    elif pos in ["l", "left"]:
        x_end, y_end = 1.02, 0.05

    y_med = 0.4

    # Define points for the arrow segments
    points = [
        (x_start, y_start),  # Start
        (x_start, y_med),  # Turn point 1
        (x_end, y_med),  # Turn point 2
        (x_end, y_end),  # End (without arrowhead)
        (x_end, 0),  # End (with arrowhead)
    ]

    linestyles = [":", "-", "-", "-", "-"]  # Differing line patterns
    alphas = [0.55, 0.8, 0.8, 1, 1]  # some transparency in the first segment
    zorders = [0, 1, 1, 1, 1]
    weight_width = 1 + 0.7 * (weight > 0.2) + 0.7 * (weight > 0.35) + 0.7 * (weight > 0.5)

    assert len(alphas) == len(points) == len(linestyles) == len(zorders)

    # Draw segments (all except final arrow head)
    for i in range(len(points) - 2):

        posA = convert_to_data_coords(ax, points[i])
        posB = convert_to_data_coords(ax, points[i + 1])

        # TODO drawn segments are separated by a small ( annoying) blank space. Fix

        arrow = FancyArrowPatch(
            posA=posA,
            posB=posB,
            arrowstyle="-",
            linewidth=weight_width,
            linestyle=linestyles[i],
            color="black",
            alpha=alphas[i],
            zorder=zorders[i],
        )
        arrow.set_clip_on(False)  # Disable clipping
        ax.add_patch(arrow)

    # Now all segemts are drawn, place text conveniently

    movesRight = 2 * ((x_end > x_start) - 0.5)
    # traslate the text to the left from the end if x_end > x_start,
    # otherwise traslate to the right
    x_text = x_end - 0.23 * movesRight

    y_text = 0.25 * y_med + 0.75 * y_end

    ax.annotate(
        text=rf"{frmt_pretty_print(pred_out, tot_digits)}$\times${weight:.2f}",  # Displaying the product of prediction and weight
        xytext=(x_text, y_text),
        xy=(x_end, y_end + 0.25),
        xycoords="axes fraction",
        textcoords="axes fraction",
        fontsize=fontsize,
        ha="center",
        va="bottom",
    )

    # Final arrowhead
    ax.annotate(
        text="",
        xytext=points[-2],
        xy=points[-1],
        xycoords="axes fraction",
        textcoords="axes fraction",
        arrowprops=dict(arrowstyle="-|>", linewidth=weight_width, mutation_scale=25, color="black"),
    )

    return ax


if __name__ == "__main__":

    # fig, ax = plt.subplots()
    # ax.set_xlim(0, 0.6)
    # ax = plot_arrow(ax, 'center', 0.2, 0.35, 1.5, fontsize=12)  # Example with 'weight' as 'w', 'pred_out' as 'p', and 'color' as 'blue'
    # plt.show()

    fig, ax = plt.subplots()
    ax.set_xlim(0, 0.70)
    ax = plot_arrow(
        ax, "c", 0.35, 0.65, fontsize=12
    )  # Example with 'weight' as 'w', 'pred_out' as 'p', and 'color' as 'blue'
    plt.show()



import warnings
import os
import numpy as np
import pandas as pd

import sklearn
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.decomposition import PCA
from sklearn.tree import _tree  # to check things like _tree.TREE_UNDEFINED

import sksurv
from sksurv.ensemble import RandomSurvivalForest
from sksurv.tree import SurvivalTree

import matplotlib as mpl
from matplotlib.colors import LinearSegmentedColormap, BoundaryNorm
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
from matplotlib.colorbar import Colorbar

from .wrapper_class import EnsembleWrapper


# def is_ci():
#     return os.environ.get("CI", "false").lower() == "true"


# def is_pytest():
#     return "PYTEST_CURRENT_TEST" in os.environ


def safe_element_to_scalar(val):
    if isinstance(val, np.ndarray) and val.shape == (1,):
        return val.item()
    return val


def get_auto_setup(y_test):

    if isinstance(y_test, np.recarray) and len(y_test.dtype.names) == 2:
        return "survival"  # single output only, for now
    else:
        y_test = np.array(y_test)
        if y_test.ndim == 1:
            unique_vals = np.unique(y_test).tolist()
            if len(unique_vals) <= 2:
                return "binary"
            else:
                return "regression"
        elif y_test.ndim == 2:  # multi-output case
            unique_vals = np.unique(y_test.ravel()).tolist()
            if len(unique_vals) <= 2:
                return "multi-label"
            else:
                return "multi-target"
        else:
            raise ValueError(f"Unexpected case, shape: {y_test.shape}")


def concatenate_helper(y_pred, y_local_pred, axis=0):

    if y_pred.shape[0] == 0:  # if still empty (no rows added)

        # Initialize final_array columns based on the first new_array
        if y_local_pred.ndim == 2:  # if output is 2D array
            y_pred = np.empty((0, y_local_pred.shape[1]))  # for (n_samples, n_outputs)
        if y_local_pred.ndim == 1:  # if 1D array instead
            y_pred = np.empty(0)  # for (n_samples,)

    # concatenate along first axis (works in any case)
    return np.concatenate((y_pred, y_local_pred), axis=axis)


def predict_helper(clf, X):
    """
    Return consistent predictions across classifiers, regressors, and wrapped models.

    INPUTS:
    - clf: a RandomForestClassifier, RandomForestRegressor,
            RandomSurvivalForest, SurvivalTree, or EnsembleWrapper instance
    - X: input data (sometimes full batch, sometimes a single sample)

    OUTPUT:
    - A NumPy array of shape:
        (n_samples,) for single-output tasks
        (n_samples, n_outputs) for multi-output
    - Or a float if X has shape (1, n_features)
    """

    def squeeze_output(y):
        """Ensure scalar for single sample, else return array."""
        y = np.array(y)
        if y.size == 1:
            return y.squeeze()
        return y

    if isinstance(clf, (RandomForestClassifier, DecisionTreeClassifier)):
        if clf.n_outputs_ == 1:
            proba = np.array(clf.predict_proba(X))
            if proba.ndim == 1:  # Handle edge case for 1D array
                result = proba
            else:
                result = (
                    proba[:, 1] if proba.shape[1] > 1 else proba[:, 0]
                )  # Return first index if only one class exists
        else:  # Multi-output classification
            result = np.array(clf.predict_proba(X))[:, :, 1].T
        return squeeze_output(result)

    elif isinstance(
        clf, (RandomForestRegressor, DecisionTreeRegressor, RandomSurvivalForest, SurvivalTree)
    ):
        result = clf.predict(X)
        return squeeze_output(result)

    elif isinstance(clf, (EnsembleWrapper, EnsembleWrapper.Estimator)):
        ypred = clf.predict(X)
        if ypred.ndim == 2 and ypred.shape[1] == 1:
            ypred = ypred.squeeze(axis=1)  # From (n_samples, 1) -> (n_samples,)
        return squeeze_output(ypred)

    else:
        raise ValueError(
            f"Tree learner '{clf.__class__.__name__}' not recognized or not implemented"
        )


def trail_pretty_digits(value, tot_digits):
    """
    Based on the order of magnitude of the value under consideration,
    leave an appropriate number of decimals. The higher the (absolute)
    value of the number, the fewer decimals.
    """
    # take a ceiling, without using the math module
    if np.abs(value) < 0.03:
        front_digits = 0  # we will use scientific notation in this case so ignore negative amounts
    else:
        front_digits = round(np.log10(np.abs(value)) + 0.5)
    front_digits = max(0, front_digits)  # we will use scientific notation in this case

    return tot_digits - front_digits


def string_to_pretty_digits(val_str, digits_single=4):

    if val_str.count(" ") > 1:  # multiple numbers stored as strings, here
        raise ValueError(f"Expected single value as a string, found: '{val_str}' instead")

    elif val_str.count(" ") == 1:
        val_str.replace(" ", "")

    if "e" in val_str:
        base, exponent = val_str.split("e")
        exponent = "e" + exponent
    else:  # leave exponent part empty:
        base, exponent = val_str, ""
    if "." in base:  # if decimal number:
        int_part, dec_part = base.split(".")
        dec_part = dec_part[: max(0, digits_single - len(int_part))]  # crop some digits from here
        cropped = f"{int_part}.{dec_part}"
    else:  # big, integer number
        cropped = base
        warnings.warn(
            f"Encountered big number with > {digits_single} digits, in non-exponenetial notation"
        )
    return cropped + exponent


def frmt_pretty_print(y_pred, digits_single=4, digits_vect=3) -> str:
    """
    Some pretty formatting from single values and np.arrays.
    Outputs values as a (single) string
    """
    assert digits_single >= 1
    assert digits_vect >= 1
    y_pred_str = None  # initialized, it's supposed to be manipulated inside the function

    y_pred = np.array(y_pred)  # ensure y_pred is a NumPy array for consistent handling

    # if 2-d, the only acceptable option is that it is a nested 1-d vector
    if isinstance(y_pred, np.ndarray) and y_pred.ndim == 2:
        if y_pred.shape[0] > 1:
            raise ValueError(
                f"Output vector must be 1d, or with shape (1,p), found shape {y_pred.shape}"
            )
        y_pred = y_pred.ravel()  # unnest the vector to get a 1-d one in case

    # case if it is a 1-d vector of shape (n,):
    if isinstance(y_pred, np.ndarray):
        if y_pred.ndim == 1 and y_pred.size > 0:  # Ensure non-empty 1D array
            count_extremes = np.sum(np.abs(y_pred) < 10 ** (-digits_vect + 2)) + np.sum(
                np.abs(y_pred) > 10 ** (digits_vect)
            )
            prop_extremes = count_extremes / len(y_pred)

            if prop_extremes < 0.5:
                y_pred_str = ", ".join(
                    f"{val:.{trail_pretty_digits(val, digits_vect)}f}" for val in y_pred
                )
            else:  # majority of numbers is very small or very big: use expon notation, drop one decimal
                y_pred_str = ", ".join(f"{val:.{digits_vect-1}e}" for val in y_pred)
        elif y_pred.size == 1:  # Handle single-value array of shape (1,)
            y_pred = y_pred.item()

    if isinstance(y_pred, (int, float)):  # case for float and for single-value array of shape (1,)
        is_extreme = np.abs(y_pred) < 10 ** (-digits_single + 2) or np.abs(y_pred) > 10 ** (
            digits_single
        )
        if not is_extreme:
            y_pred_str = f"{y_pred:.{trail_pretty_digits(y_pred, digits_single)}f}"
        else:  # very small or very big: use expon notation, drop one decimal
            y_pred_str = f"{y_pred:.{digits_single-1}e}"

    if isinstance(y_pred, str):
        warnings.warn(
            "y_pred is already a string. You might be calling the formatting function twice"
        )
    if y_pred_str is None:
        # then none of the previous cases was encountered (ideally use elif but it's tricky)
        raise ValueError(
            f"Format for y_pred not recognized, or its shape is unusual."
            f"\nFound {type(y_pred)}: {y_pred}"
        )

    return y_pred_str


def return_partial_preds(clf_i):

    if isinstance(clf_i, sklearn.tree.DecisionTreeClassifier) and clf_i.n_outputs_ == 1:
        partials = clf_i.tree_.value[:, 0, :]  # now take average
        partial_preds = partials[:, 1] / (partials[:, 0] + partials[:, 1])

    elif isinstance(clf_i, sklearn.tree.DecisionTreeClassifier) and clf_i.n_outputs_ > 1:
        partials = clf_i.tree_.value
        partial_preds = partials[:, :, 1] / (partials[:, :, 0] + partials[:, :, 1])

    elif isinstance(clf_i, sklearn.tree.DecisionTreeRegressor) and clf_i.n_outputs_ == 1:
        partial_preds = clf_i.tree_.value.ravel()  # DOUBLE CHECK!

    elif isinstance(clf_i, sklearn.tree.DecisionTreeRegressor) and clf_i.n_outputs_ > 1:
        partial_preds = clf_i.tree_.value.squeeze(axis=-1)  # (n,p,1) to (n,p)

    elif isinstance(clf_i, SurvivalTree):
        # clf_i.tree_.value: np array of [node, time, [H(node), S(node)]]
        #                                              ^idx 0   ^idx 1
        # We imitate the .predict function of the SurvivalTree:
        partial_preds = np.sum(clf_i.tree_.value[:, clf_i.is_event_time_, 0], axis=1)

    elif isinstance(clf_i, EnsembleWrapper.Estimator):
        if clf_i.n_outputs_ == 1:
            partial_preds = clf_i.tree_.value.ravel()  # .ravel seems to do the needed formatting
        else:
            partial_preds = clf_i.tree_.value
    else:
        raise ValueError("Tree learner not recognized, or not implemented")

    return partial_preds


def used_feature_set(clf_i, feature_names, sample):

    unique_features = []
    # tested for RandomForestClassifier and EnsembleWrapper (binary set-up)

    node_indicator_csr = clf_i.decision_path(sample.values)  # sparse matrix (1, n_nodes)
    feature_idx_per_node = clf_i.tree_.feature  # array (n_nodes, )

    node_index = node_indicator_csr.indices[
        node_indicator_csr.indptr[0] : node_indicator_csr.indptr[1]
    ]  # csr matrix formatted in this way

    for node_id in node_index[:-1]:  # internal nodes (exclude leaf)
        feature_node_id = feature_idx_per_node[node_id]
        feature_name = feature_names[feature_node_id]
        if feature_name not in unique_features:  # add element to list if not in there yet
            unique_features.append(feature_name)

    return unique_features


def colormap_from_str(colormap):
    """
    Function for the user to customize the colormap.
    Accepts a matplotlib colormap object or a string.
    """

    if colormap is None:
        # Default: use 'RdYlBu_r'
        cmap_output = mpl.colormaps["RdYlBu_r"]

    elif isinstance(colormap, LinearSegmentedColormap):
        cmap_output = colormap

    elif isinstance(colormap, str):
        if colormap not in mpl.colormaps:
            raise ValueError(
                f'Provided string "{colormap}" is not a recognized colormap.\n'
                f"Check available names via list(matplotlib.colormaps)."
            )
        cmap_output = mpl.colormaps[colormap]

    else:
        raise ValueError(
            f"Provided colormap must be either a LinearSegmentedColormap or a recognized string.\n"
            f"Got {type(colormap)} instead."
        )

    return cmap_output


def rule_print_inline(clf_i, sample, weight=None, max_features_print=12):
    """sample is a pd.Series or a single-row pd.DataFrame??"""
    ## consider treating it as a numpy array
    if isinstance(sample, np.ndarray):
        sample = pd.DataFrame(sample)
        sample.columns = [f"X_{i}" for i in range(len(sample.columns))]

    # node_indicator = clf_i.decision_path(sample)
    node_indicator_csr = clf_i.decision_path(sample.values)
    # node_weights = clf_i.tree_.n_node_samples/(clf_i.tree_.n_node_samples[0])
    children_left = clf_i.tree_.children_left
    children_right = clf_i.tree_.children_right
    feature = clf_i.tree_.feature
    threshold = clf_i.tree_.threshold

    is_traversed_node = node_indicator_csr.indices[
        node_indicator_csr.indptr[0] : node_indicator_csr.indptr[1]
    ]  # csr matrix formatted in this way

    unique_features = used_feature_set(clf_i, sample.columns, sample)

    # Print only the relevant features with max 4 digits
    # take care of selection of sample columns so that it stays as pd.DataFrame:
    unique_features_formatted = sample[unique_features].apply(
        lambda col: col.map(lambda x: frmt_pretty_print(x, 4))
    )

    if len(unique_features) <= max_features_print:
        print("#" * 22, "   SAMPLE   ", "#" * 22)
        print(unique_features_formatted.to_string(col_space=4))
        print("#" * 58)
    else:
        print("#" * 58)
        print("Too many features are used in the extracted rules, therefore we", end="")
        print("skip the printing. \n Increase the max_features_print parameter in case")
        print("#" * 58)

    partial_preds = return_partial_preds(clf_i)

    if weight is None:
        print(f"Baseline prediction: {frmt_pretty_print(partial_preds[0])}")
    else:
        print(
            f"Baseline prediction: {frmt_pretty_print(partial_preds[0])} \t (weight = {weight:.2f})"
        )

    for node_id in is_traversed_node[:-1]:  # internal nodes (exclude leaf)
        # continue to the next node if it is a leaf node

        # check if value of the split feature for sample 0 is below threshold
        if sample.values[0, feature[node_id]] <= threshold[node_id]:
            threshold_sign = "<="
            next_child = children_left[node_id]
        else:
            threshold_sign = "> "
            next_child = children_right[node_id]

        print(
            f"node {node_id:3}: "
            f"{sample.columns[feature[node_id]]:>8} {threshold_sign} {threshold[node_id]:5>.2f} "
            f"{'(' + sample.columns[feature[node_id]]:>9} = {sample.values[0, feature[node_id]]:5>.2f})"
            f"  -->  {frmt_pretty_print(partial_preds[next_child])}"
        )

    print(
        f"leaf {is_traversed_node[-1]:4}: predicts: "
        f"{frmt_pretty_print(partial_preds[is_traversed_node[-1]])}"
    )


def rule_to_file(clf_i, sample, rule_weight, max_features_print, f):

    # leaf_print = predict_helper(clf_i, sample.values)
    partial_preds = return_partial_preds(clf_i)

    feature_names = sample.columns

    def recurse_print(node, depth, tree_, sample, feature_names, is_traversed_node, f):
        indent = "  " * depth

        if tree_.feature[node] != _tree.TREE_UNDEFINED:  # if feature is not undefined != -2
            name = feature_names[tree_.feature[node]]
            threshold = tree_.threshold[node]

            if is_traversed_node[node] == 1 and sample[tree_.feature[node]] <= threshold:
                is_traversed_node[node] = 0  # otherwise it will keep printing this rulesplit
                child_node = tree_.children_left[node]
                threshold_str = frmt_pretty_print(threshold, 6)
                partial_preds_str = frmt_pretty_print(partial_preds[child_node], 6)
                f.write(
                    f"node.{node:4}: {indent} {name} <= {threshold_str}  --> {partial_preds_str}\n"
                )

                recurse_print(
                    child_node, depth + 1, tree_, sample, feature_names, is_traversed_node, f
                )

            if is_traversed_node[node] == 1 and sample[tree_.feature[node]] > threshold:
                is_traversed_node[node] = 0
                child_node = tree_.children_right[node]
                threshold_str = frmt_pretty_print(threshold, 6)
                partial_preds_str = frmt_pretty_print(partial_preds[child_node], 6)
                f.write(
                    f"node.{node:4}: {indent} {name} >  {threshold_str}  --> {partial_preds_str}\n"
                )

                recurse_print(
                    child_node, depth + 1, tree_, sample, feature_names, is_traversed_node, f
                )

        else:  # if feature split is == _tree.TREE_UNDEFINED (that is, -2), then we are in a leaf
            if is_traversed_node[node] == 1:
                partial_preds_str = frmt_pretty_print(partial_preds[node], 6)
                f.write(f"leaf.{node:4}: {indent}returns {partial_preds_str}\n")

    tree_structure = clf_i.tree_
    unique_features = used_feature_set(clf_i, feature_names, sample)

    # Take care of selection of sample columns so that
    # it stays in the pd.DataFrame format. This works:
    # store values with at most 6 digits in the file ( whjta baout col_space = 5 though?)
    unique_features_formatted = sample[unique_features].apply(
        lambda col: col.map(lambda x: frmt_pretty_print(x, 6))
    )

    if len(unique_features) <= max_features_print:
        f.write("#" * 24 + "  SAMPLE  " + "#" * 24 + "\n")
        f.write(unique_features_formatted.to_string(col_space=5) + "\n")
    else:
        f.write("#" * 58 + "\n")

    f.write("#" * 18 + f"   RULE WEIGHT: {rule_weight:.2f} " + "#" * 18 + "\n")
    f.write(f"Baseline prediction: {frmt_pretty_print(partial_preds[0], 6)}\n")

    is_traversed_node = clf_i.decision_path(sample.values).toarray()[0]
    sample = sample.to_numpy().reshape(-1)  # from single column to single line
    ## and here most of the printing is done (recursive calls)
    recurse_print(
        0, 0, tree_structure, sample, feature_names, is_traversed_node, f
    )  # feature_name list missing?


def custom_axes_limit(bunch_min_value, bunch_max_value, force_in, is_binary):

    if force_in is None:
        force_in = np.nan  # so that it can be ignored in later computations (min, max)

    v_min = min(bunch_min_value, force_in)
    v_max = max(bunch_max_value, force_in)

    if is_binary:
        # combat counterintuitive colouring when predictions are very confident
        # if all predicitons are very low, they are all mapped in the lower part of the ColorMap
        v_min = min(v_min, 0.7)  # v_min never above 0.7
        # if all predicitons are very high, they are all mapped in the upper part of the ColorMap
        v_max = max(v_max, 0.3)  # v_max never below 0.3

    # add a bit of extra spacing on the extremes, to avoid the case v_min = v_max
    # still possible if e.g. v_min ~ v_max ~ 0.5
    v_min = v_min - (v_max - v_min) * 0.02
    v_max = v_max + (v_max - v_min) * 0.02 + 0.005

    return v_min, v_max


def custom_formatter(x, pos):  # pos paramter to comply with expected signature
    """
    Custom formatter function, useful for the colorabar
    """
    if 1e-2 <= np.abs(x) < 1:
        return f"{x:.2f}"  # 2 decimal digits for numbers between -1 and 1, with asb value >= 0.01
    elif 1 <= np.abs(x) < 10:
        return (
            f"{x:.1f}"  # 1 decimal digit for numbers with 1 significant digit before decimal point
        )
    elif 10 <= np.abs(x) < 100:
        return f"{x:.0f}"  # round to nearest integer for numbers with 2 digits before decimal point
    else:  # np.abs(x) < 1e-2 or np.abs(x) > 100 (very big or very small numbers)
        return f"{x:.1e}"  # Scientific notation with 2 significant digits (x.y1e__)

    ## LocalMethod inputs: plot_data_bunch, plot_kmeans, tuned_method, self.clf.n_outputs_


def plot_preselected_trees(
    plot_data_bunch,
    kmeans,
    tuned_method,
    base_font_size=12,
    show_ax_ticks="auto",
    colormap=None,
    alpha_dots=0.5,
):

    small_size = 40
    big_size = 220

    if show_ax_ticks == "auto":
        show_ax_ticks = False if base_font_size > 15 else True

    # PCA to 2 dimensions for projected trees
    # (original proj dimension can be > 2)
    PCA_fitted = PCA(n_components=2).fit(plot_data_bunch.proj_data)
    plottable_data = PCA_fitted.transform(plot_data_bunch.proj_data)  # (lambda,2)

    centers = PCA_fitted.transform(kmeans.cluster_centers_)
    class_memb = kmeans.labels_

    custom_gridspec = {"width_ratios": [3, 0.2, 3, 0.2]}

    fig = plt.figure(figsize=(10, 4.5))
    axes = fig.subplots(1, 4, gridspec_kw=custom_gridspec)
    # fig.subplots_adjust(top=0.85)
    # fig.tight_layout()

    # conditional sizes for trees and candidate trees:
    is_final_candidate = [
        plot_data_bunch.index[i] in tuned_method.final_trees_idx
        for i in range(len(plot_data_bunch.index))
    ]

    #####   LEFT PLOT (cluster memberships)   #####

    for i, txt in enumerate(centers):  # plot cluster centers
        axes[0].annotate(
            i + 1,
            (txt[0], txt[1]),  # old .annotate(i+1, centers[i,0], centers[i,1]),
            bbox={"boxstyle": "circle", "color": "grey", "alpha": 0.6},
        )

    x_normal = plottable_data[:, 0][[not x for x in is_final_candidate]]
    y_normal = plottable_data[:, 1][[not x for x in is_final_candidate]]
    color_normal = class_memb[[not x for x in is_final_candidate]]

    x_candidate = plottable_data[:, 0][is_final_candidate]
    y_candidate = plottable_data[:, 1][is_final_candidate]
    color_candidate = class_memb[is_final_candidate]

    axes[0].scatter(
        x_normal,
        y_normal,
        c=color_normal,
        cmap=None,
        s=small_size,
        marker="o",
        edgecolors=(1, 1, 1, alpha_dots),
    )

    axes[0].scatter(
        x_candidate,
        y_candidate,
        c=color_candidate,
        cmap=None,
        s=big_size,
        marker="*",
        edgecolors="black",
    )

    axes[0].set_xlabel("PC1", fontdict={"fontsize": base_font_size - 1})
    axes[0].set_ylabel("PC2", fontdict={"fontsize": base_font_size - 1})

    axes[0].axis("equal")  # is it even a good idea? We will see
    axes[0].set_title("Cluster membership", fontdict={"fontsize": base_font_size + 1})

    # create the map for segmented colorbar (axes[1]: left colorbar)
    cmap = plt.colormaps["viridis"]
    cmaplist = [cmap(i) for i in range(cmap.N)]
    cmap_left = LinearSegmentedColormap.from_list("Custom cmap", cmaplist, cmap.N)

    # define the bins and normalize
    freqs = np.bincount(class_memb)
    if np.min(freqs) == 0:
        raise KeyError(
            "There are empty clusters, the scatter and colorbar would differ in color shade"
        )
    norm_bins = list(np.cumsum(freqs))
    norm_bins.insert(0, 0)

    if len(norm_bins) == 2:  # color gradient is off, add artificial bin
        # this will create an empty artificial cluster later on, that will be dropped
        norm_bins.insert(-1, norm_bins[1])

    # scatterplot color does not scale correctly if there are empty classes!
    # transform list to array (ticks location needs arithmentic computation)
    norm_bins = np.array(norm_bins)

    # create label names
    labels = []
    for i in np.unique(class_memb):
        labels.append("cl.{:d}".format(i + 1))

    # normalizing color, prepare ticks, labels
    norm = BoundaryNorm(norm_bins, cmap_left.N)
    tickz = norm_bins[:-1] + (norm_bins[1:] - norm_bins[:-1]) / 2

    if tickz.max() == norm_bins.max():  # artificial empty cluster somewhere: drop
        tickz = tickz[:-1]  # drop last tick at top of colorbar

    # colorab on axis 2 out of 4.
    cb = Colorbar(
        axes[1],
        cmap=cmap_left,
        norm=norm,
        spacing="proportional",
        ticks=list(tickz),
        boundaries=list(norm_bins),
        format="%1i",
    )
    # label="cluster membership")
    cb.ax.set_yticklabels(labels)  # vertically oriented colorbar
    cb.ax.tick_params(labelsize=base_font_size - 1)  # ticks font size
    axes[1].yaxis.set_ticks_position("left")

    # User can customize the colorbar
    cmap_right = colormap_from_str(colormap)

    #####   RIGHT PLOT (predictions or losses)  #####

    # PREDICTIONS when single class output (SurvivalTree included)
    if tuned_method.clf.n_outputs_ == 1 or isinstance(
        tuned_method.clf, RandomSurvivalForest
    ):  # single output, color on predictions

        ### right figure scatterplot here (axes[2] and axes[3]):

        is_binary = False
        if isinstance(tuned_method.clf, sklearn.ensemble.RandomForestClassifier):
            is_binary = tuned_method.clf.n_outputs_ == 1
        elif isinstance(tuned_method.clf, EnsembleWrapper):
            is_binary = (tuned_method.clf.n_outputs_ == 1) and (
                tuned_method.clf.ensemble_class == "RandomForestClassifier"
            )

        v_min, v_max = custom_axes_limit(
            np.array(plot_data_bunch.pred).min(),
            np.array(plot_data_bunch.pred).max(),
            force_in=plot_data_bunch.rf_pred,
            is_binary=is_binary,
        )

        norm_preds = BoundaryNorm(np.linspace(v_min, v_max, 256), cmap_right.N)

        color_indeces = np.zeros(len(plot_data_bunch.pred))  # length = n_trees

        for i, bunch_pred in enumerate(plot_data_bunch.pred):
            # count number of values in norm_preds.boundaries that are less than the prediction
            color_indeces[i] = np.argmin([thresh <= bunch_pred for thresh in norm_preds.boundaries])

        # format as integers, for list comprehension
        color_indeces = [int(x + 0.1) for x in color_indeces]

        real_colors = np.array([cmap_right(idx) for idx in color_indeces])

        axes[2].scatter(
            x_normal,
            y_normal,
            c=real_colors[[not x for x in is_final_candidate]],
            s=small_size,  # cmap=cmap_right,
            marker="o",
            edgecolors=(1, 1, 1, alpha_dots),
        )

        axes[2].scatter(
            x_candidate,
            y_candidate,
            c=real_colors[is_final_candidate],
            s=big_size,  # cmap=cmap_right,
            marker="*",
            edgecolors="black",
        )

        axes[2].set_xlabel("PC1", fontdict={"fontsize": base_font_size - 1})
        axes[2].yaxis.set_label_position("right")
        axes[2].set_ylabel("PC2", fontdict={"fontsize": base_font_size - 1})
        # axes[2].yaxis.tick_right()
        axes[2].axis("equal")
        axes[2].set_title("Rule-path predictions", fontdict={"fontsize": base_font_size + 1})

        # add color bar to the side
        pred_tick = np.round(float(tuned_method.local_prediction().item()), 3)

        cb2 = Colorbar(
            axes[3],
            cmap=cmap_right,
            norm=norm_preds,
            format=FuncFormatter(custom_formatter),
            label="predicted: " + str(pred_tick),
        )

        ## add to colorbar a line corresponding to Bellatrex prediction

        pred_lines = [float(x) for x in plot_data_bunch.pred]

        cb2.ax.plot([0, 1], [pred_lines] * 2, color="grey", linewidth=1)
        cb2.ax.plot([0.02, 0.98], [pred_tick] * 2, color="black", linewidth=2.5, marker="P")

        if isinstance(tuned_method.clf, sksurv.ensemble.RandomSurvivalForest):
            cb2.set_label("Cumul. Hazard: " + str(pred_tick), size=base_font_size - 3)

        elif isinstance(tuned_method.clf, sklearn.ensemble.RandomForestClassifier):
            cb2.set_label("Pred. Prob:" + str(pred_tick), size=base_font_size - 3)
        elif isinstance(tuned_method.clf, sklearn.ensemble.RandomForestRegressor):
            cb2.set_label("Pred. value:" + str(pred_tick), size=base_font_size - 3)

        elif isinstance(tuned_method.clf, EnsembleWrapper):
            if tuned_method.clf.ensemble_class == "RandomSurvivalForest":
                cb2.set_label("Cumul. Hazard: " + str(pred_tick), size=base_font_size - 3)

            elif tuned_method.clf.ensemble_class == "RandomForestClassifier":
                cb2.set_label("Pred. prob:" + str(pred_tick), size=base_font_size - 3)
            elif tuned_method.clf.ensemble_class == "RandomForestRegressor":
                cb2.set_label("Pred. value:" + str(pred_tick), size=base_font_size - 3)
        else:
            raise ValueError(f"Model not recognized: {tuned_method.clf}")

    # LOSS  when multi-output predictions: plot distance from RF preds. The lower the better (blue)
    else:
        # adds padding betwwen v_min and v_max in case they coincide
        v_min, v_max = custom_axes_limit(
            np.array(plot_data_bunch.loss).min(),
            np.array(plot_data_bunch.loss).max(),
            force_in=np.nan,
            is_binary=False,
        )

        norm_preds = BoundaryNorm(np.linspace(v_min, v_max, 256), cmap.N)

        final_candidate_loss = np.array(plot_data_bunch.loss)[is_final_candidate]
        normal_rule_loss = np.array(plot_data_bunch.loss)[[not x for x in is_final_candidate]]

        axes[2].scatter(
            x_normal,
            y_normal,
            c=normal_rule_loss,
            cmap=cmap_right,
            norm=norm_preds,
            s=small_size,
            marker="o",
            edgecolors=(1, 1, 1, alpha_dots),
        )

        axes[2].scatter(
            x_candidate,
            y_candidate,
            c=final_candidate_loss,
            cmap=cmap_right,
            norm=norm_preds,
            s=big_size,
            marker="*",
            edgecolors="black",
        )

        axes[2].set_xlabel("PC1", fontdict={"fontsize": base_font_size - 1})
        axes[2].yaxis.set_label_position("right")
        axes[2].set_ylabel("PC2", fontdict={"fontsize": base_font_size - 1})
        # axes[2].yaxis.tick_right()
        axes[2].axis("equal")
        axes[2].set_title("Rule-path predictions", fontdict={"fontsize": base_font_size + 1})

        cb2 = Colorbar(
            axes[3],
            cmap=cmap_right,
            norm=norm_preds,
            label=str(tuned_method.fidelity_measure) + " loss",
        )
        cb2.ax.plot([0, 1], [plot_data_bunch.loss] * 2, color="grey", linewidth=1)

    # end of single-target vs multi-target case (de-indent)
    ticks_to_plot = axes[3].get_yticks()

    if np.abs(np.min(ticks_to_plot)) < 1e-3 and np.abs(np.max(ticks_to_plot)) > 1e-2:
        min_index = np.argmin(ticks_to_plot)
        ticks_to_plot[min_index] = 0
        axes[3].set_yticks(ticks_to_plot)

    axes[3].yaxis.set_major_formatter(FuncFormatter(custom_formatter))
    axes[3].minorticks_off()

    cb2.ax.tick_params(labelsize=base_font_size - 3)  # ticks font size

    if show_ax_ticks is False:
        axes[0].set_xticklabels([])
        axes[0].set_yticklabels([])
        axes[2].set_xticklabels([])
        axes[2].set_yticklabels([])

    return fig, axes


#### LEGACY CODE HERE? COMMENTED OUT, CONSIDER THROWING IT AWAY ####

# def rule_to_code(clf_i, traversed_nodes, sample, full_save_name):

#     leaf_print = predict_helper(clf_i, sample.values)

#     tree_ = clf_i.tree_
#     feature_names = sample.columns  # it's a pd.DataFrame by now

#     feature_name = [
#         feature_names[i] if i != _tree.TREE_UNDEFINED else "undefined!" for i in tree_.feature
#     ]

#     intervals = {feat: [-np.inf, np.inf] for feat in feature_names}

#     if full_save_name is not None:
#         with open(full_save_name, "w+", encoding="utf-8") as f:
#             f.write("###### SAMPLE to explain ######\n")

#             for i, k in zip(feature_names, range(len(feature_names))):
#                 f.write(f"{str(i):13}: {str(sample[k]):7} \n")

#             f.write("\n###############################\n")

#             sample = sample.to_numpy().reshape(-1)  # from single column to single line

#             def recurse(node, depth, sample, intervals):
#                 indent = "  " * depth
#                 if tree_.feature[node] != _tree.TREE_UNDEFINED:  # if feature is not undefined (??)
#                     name = feature_name[node]
#                     threshold = tree_.threshold[node]
#                     if traversed_nodes[node] == 1 and sample[tree_.feature[node]] <= threshold:
#                         intervals[name][1] = threshold  # reduce feature upper bound
#                         traversed_nodes[node] = 0
#                         f.write(f"node.{name}:{indent} if {name} <= {threshold}\n")

#                     recurse(tree_.children_left[node], depth + 1, sample, intervals)

#                     if traversed_nodes[node] == 1 and sample[tree_.feature[node]] > threshold:
#                         intervals[name][0] = threshold  # increase feature lower bound
#                         traversed_nodes[node] = 0
#                         f.write(f"node {name}:{indent} if {name} > {threshold}\n")
#                     recurse(tree_.children_right[node], depth + 1, sample, intervals)
#                 else:  # it is undefined, it is therefore a leaf (?)
#                     if traversed_nodes[node] == 1:
#                         # print("leafnode.{}: {}return {}".format(node, indent, leaf_print2)) #tree_.value[node].ravel()
#                         f.write(f"leafnode.{name}:{indent} returns {leaf_print}\n")
#                         f.write(f"predicted:{leaf_print}\n")

#             recurse(0, 1, sample, intervals)
#             f.close()


# def rule_to_code_and_intervals(clf_i, traversed_nodes, sample, feature_names, full_save_name):

#     leaf_print = predict_helper(clf_i, sample)

#     tree_ = clf_i.tree_
#     feature_name = [feature_names[i] if i != -2 else "undefined!" for i in tree_.feature]

#     intervals = {feat: [-np.inf, np.inf] for feat in feature_names}

#     if full_save_name is not None:
#         with open(full_save_name, "w+", encoding="utf-8") as f:
#             f.write("###### SAMPLE to explain ######\n")

#             for i, k in zip(feature_names, range(len(feature_names))):
#                 f.write(f"{str(i):13}: {str(sample[k]):7} \n")

#             f.write("\n###############################\n")

#             sample = sample.to_numpy().reshape(-1)  # from single column to single line

#             def recurse(node, depth, sample, intervals):
#                 indent = "  " * depth
#                 if tree_.feature[node] != -2:
#                     name = feature_name[node]
#                     threshold = tree_.threshold[node]
#                     if traversed_nodes[node] == 1 and sample[tree_.feature[node]] <= threshold:
#                         intervals[name][1] = threshold  # reduce feature upper bound
#                         traversed_nodes[node] = 0
#                         f.write(f"node.{node}:{indent} if {name} <= {threshold}\n")

#                     recurse(tree_.children_left[node], depth + 1, sample, intervals)

#                     if traversed_nodes[node] == 1 and sample[tree_.feature[node]] > threshold:
#                         intervals[name][0] = threshold  # increase feature lower bound
#                         traversed_nodes[node] = 0
#                         f.write(f"node.{node}:{indent} if {name} > {threshold}\n")
#                     recurse(tree_.children_right[node], depth + 1, sample, intervals)
#                 else:  # it is undefined, it is therefore a leaf (?)
#                     if traversed_nodes[node] == 1:
#                         f.write(f"leafnode.{node}:{indent} return {leaf_print}\n")
#                         f.write(f"predicted:{leaf_print}\n")

#             recurse(0, 1, sample, intervals)
#             f.close()

#     if full_save_name is not None:
#         with open(
#             full_save_name.split(".")[0] + "-simplif." + full_save_name.split(".")[-1],
#             "w+",
#             encoding="utf-8",
#         ) as f:
#             f.write("###### SAMPLE to explain ######\n")

#             for i, k in zip(feature_names, range(len(feature_names))):
#                 f.write(f"{str(i):10}: {str(sample[k]):7}\n")

#             f.write("\n###### final intervals ########\n")

#             for item in intervals:
#                 if intervals[item][0] != -np.inf or intervals[item][1] != np.inf:
#                     f.write(
#                         f"{intervals[item][0]:6} < {str(item).center(8)} "
#                         f"<= {intervals[item][1]:6} \n"
#                     )
#             f.close()

#             with open(
#                 full_save_name, encoding="utf-8"
#             ) as f:  # printing tree-rule structure on console
#                 print(f.read())

#             print(
#                 "###############################"
#             )  # split between tree rule print and leaf interval representation

#             with open(
#                 full_save_name.split(".")[0] + "-simplif." + full_save_name.split(".")[-1],
#                 encoding="utf-8",
#             ) as f:
#                 print(f.read())  # printing (simplified) leaf structure on console



import pytest
import numpy as np
from bellatrex.utilities import (
    get_auto_setup,
    concatenate_helper,
    trail_pretty_digits,
    string_to_pretty_digits,
    frmt_pretty_print,
)


def test_get_auto_setup():
    # Test binary classification
    y_test = np.array([0, 1, 0, 1])
    assert get_auto_setup(y_test) == "binary"

    # Test regression
    y_test = np.array([1.2, 3.4, 5.6])
    assert get_auto_setup(y_test) == "regression"

    # Test multi-label classification
    y_test = np.array([[0, 1], [1, 0], [0, 1]])
    assert get_auto_setup(y_test) == "multi-label"

    # Test multi-target regression
    y_test = np.array([[1.2, 3.4], [5.6, 7.8]])
    assert get_auto_setup(y_test) == "multi-target"


def test_concatenate_helper():
    y_pred = np.empty((0, 2))
    y_local_pred = np.array([[1, 2], [3, 4]])
    result = concatenate_helper(y_pred, y_local_pred)
    assert result.shape == (2, 2)
    assert np.array_equal(result, y_local_pred)


def test_trail_pretty_digits():
    assert trail_pretty_digits(0.00123, 4) == 4  # Updated expected output
    assert trail_pretty_digits(123.456, 4) == 1


def test_string_to_pretty_digits():
    assert string_to_pretty_digits("123.456", digits_single=4) == "123.4"
    assert string_to_pretty_digits("1.23e-4", digits_single=4) == "1.23e-4"


def test_frmt_pretty_print():
    assert frmt_pretty_print(np.array([0.123, 0.456]), digits_vect=3) == "0.123, 0.456"
    assert frmt_pretty_print(123.456, digits_single=4) == "123.5"



import os
import pytest
import numpy as np
import pandas as pd

try:  # TODO: Paths to be updated: this workaround makes tests work across different setups.
    from app.bellatrex.bellatrex_explain import BellatrexExplain
except ImportError:
    from bellatrex.bellatrex_explain import BellatrexExplain
from sklearn.ensemble import RandomForestClassifier
from sksurv.ensemble import RandomSurvivalForest


@pytest.fixture
def mock_clf():
    return RandomForestClassifier(n_estimators=10, random_state=0)


@pytest.fixture
def mock_survival_clf():
    return RandomSurvivalForest(n_estimators=10, random_state=0)


@pytest.fixture
def mock_data():
    X = pd.DataFrame(np.random.rand(100, 5), columns=[f"feature_{i}" for i in range(5)])
    y = np.random.randint(0, 2, size=100)  # random binary target
    return X, y


@pytest.fixture
def mock_survival_data():
    X = pd.DataFrame(np.random.rand(100, 5), columns=[f"feature_{i}" for i in range(5)])

    times = np.random.exponential(scale=5, size=100).astype(float)
    status = np.random.binomial(1, 0.7, size=100).astype(bool)
    # order of columns is: status, times
    y = np.rec.fromarrays([status, times], dtype=[("event", "?"), ("time", "f8")])

    return X, y


def test_bellatrex_explain_fit(mock_clf, mock_data):
    X, y = mock_data
    explainer = BellatrexExplain(mock_clf, verbose=1)
    explainer.fit(X, y)
    assert explainer.is_fitted() is True


# def test_bellatrex_explain_explain(mock_clf, mock_data):
#     X, y = mock_data
#     explainer = BellatrexExplain(mock_clf, verbose=0)
#     explainer.fit(X, y)
#     explanation = explainer.explain(X, idx=0)
#     assert explanation is not None


def test_bellatrex_explain_plot_overview(mock_clf, mock_data):
    X, y = mock_data
    explainer = BellatrexExplain(mock_clf, verbose=1)
    explainer.fit(X, y)
    fig, axes = explainer.explain(X, idx=0).plot_overview(show=False)
    # fig, axes = explainer.plot_overview(show=False)
    assert fig is not None
    assert axes is not None


@pytest.mark.xfail(raises=NotImplementedError, reason="Not implemented yet")
def test_predict_survival_curve(mock_survival_clf, mock_survival_data):
    X, y = mock_survival_data
    explainer = BellatrexExplain(mock_survival_clf, verbose=1, set_up="survival")
    explainer.fit(X, y)
    survival_curve = explainer.predict_survival_curve(X, 0)  # pylint: disable=E1111

    assert survival_curve is not None
    assert isinstance(survival_curve, pd.DataFrame)


@pytest.mark.xfail(raises=NotImplementedError, reason="Not implemented yet")
def test_predict_median_surv_time(mock_survival_clf, mock_survival_data):
    X, y = mock_survival_data
    explainer = BellatrexExplain(mock_survival_clf, verbose=1, set_up="survival")
    explainer.fit(X, y)
    survival_curve = explainer.predict_median_surv_time(X, 0)  # pylint: disable=E1111
    assert survival_curve is not None
    assert isinstance(survival_curve, pd.DataFrame)


# BIG DUMMY TREE TO MAKE EVERYTHING WORK.
# Better refactgor code than go through this pain.
# class DummyTree(dict):
#     def __init__(self):
#         super().__init__()
#         self.value = [[0]]
#         self.n_outputs_ = 1
#         self["feature_names_in_"] = ["f1", "f2"]
#         self["n_features_in_"] = 2

#     def get(self, *a, **k):
#         return self.value


# def test_is_fitted_dict():
#     # Should wrap dict in EnsembleWrapper and return True

#     dict_trees = {"trees": [DummyTree(), DummyTree(), DummyTree()]}
#     explainer = BellatrexExplain(dict_trees)
#     assert explainer.is_fitted() is True


# def test_is_fitted_ensemblewrapper():
#     from bellatrex.wrapper_class import EnsembleWrapper

#     dict_trees = {"trees": [DummyTree(), DummyTree(), DummyTree()]}
#     ew = EnsembleWrapper(dict_trees)
#     explainer = BellatrexExplain(ew)
#     assert explainer.is_fitted() is True


# def test_fit_force_refit_verbose(monkeypatch, mock_clf, mock_data):
#     X, y = mock_data
#     explainer = BellatrexExplain(mock_clf, force_refit=True, verbose=2)
#     called = {}

#     def fake_fit(X_, y_, n_jobs):
#         called["fit"] = True

#     monkeypatch.setattr(mock_clf, "fit", fake_fit)
#     explainer.fit(X, y)
#     assert called["fit"]


# def test_explain_verbose(monkeypatch, mock_clf, mock_data):
#     X, y = mock_data
#     explainer = BellatrexExplain(mock_clf, verbose=5)
#     explainer.fit(X, y)

#     # Patch TreeExtraction to avoid heavy computation
#     class DummyTE:
#         def __init__(self, *a, **k):
#             pass

#         def set_params(self, **params):
#             return self

#         def main_fit(self):
#             class Dummy:
#                 final_trees_idx = [0]
#                 cluster_sizes = [1]
#                 score = lambda self, *a, **k: 1.0

#             return Dummy()

#     monkeypatch.setattr("app.bellatrex.bellatrex_explain.TreeExtraction", DummyTE)
#     explainer.explain(X, idx=0)
#     assert hasattr(explainer, "tuned_method")


def test_create_rules_txt_file(monkeypatch, mock_clf, mock_data):
    X, y = mock_data
    explainer = BellatrexExplain(mock_clf)
    explainer.fit(X, y)

    class DummyModel:
        final_trees_idx = [0]
        cluster_sizes = [1]
        sample = X.iloc[[0]]

    explainer.tuned_method = DummyModel()
    explainer.sample = X.iloc[[0]]
    explainer.sample_iloc = 0
    explainer.surrogate_pred_str = "0.0"
    explainer.clf = mock_clf
    # Monkeypatch rule_to_file and read_rules to avoid file IO
    # should paths include app.bellatrex or just bellatrex? TBD and TB fixed
    monkeypatch.setattr("bellatrex.utilities.rule_to_file", lambda *a, **k: None)
    monkeypatch.setattr("bellatrex.visualization.read_rules", lambda **k: ([1], [1], [1], [1], [1]))
    monkeypatch.setattr("bellatrex.visualization_extra._input_validation", lambda *a, **k: None)
    out_file, file_extra = explainer.create_rules_txt(
        out_dir="explanations-output", out_file="testing_rules.txt"
    )
    assert os.path.exists(out_file)
    assert os.path.exists(file_extra)

    os.remove(out_file)
    os.remove(file_extra)
